from AutoRPE.UtilsRPE.BasicFunctions import close_brackets
from AutoRPE.UtilsRPE.CallManager import find_argument_intervals
from AutoRPE.UtilsRPE.Finder import find_file_and_line_of_allocation, \
    find_file_and_line_of_namelist_reading, find_first_use

from AutoRPE.UtilsRPE.Getter import retrieve_variable

import warnings
import re


def insert_variable_precision_specification(variable, sourcefile, insertion_line, condition=None):
    if condition is None:
        condition = ""
    # Add precision specification
    new_line = condition + "    %s%%sbits = emulator_variable_precisions(%i)" % (variable.name, variable.id)
    sourcefile.lines.insert(insertion_line, new_line)

    # Add variable usage specification
    new_line = condition + "    variable_is_used(%i) = .true." % variable.id
    sourcefile.lines.insert(insertion_line, new_line)

    sourcefile.rebuild_text()


def insert_variable_usage(variable, sourcefile, insertion_line, condition=None):
    if condition is None:
        condition = ""

    # Add variable usage specification
    new_line = condition + "    variable_is_used(%i) = .true." % variable.id
    sourcefile.lines.insert(insertion_line, new_line)

    sourcefile.rebuild_text()


def insert_variable_truncation(variable, sourcefile, insertion_line, condition=None):
    if condition is None:
        condition = ""
    new_line = condition + "    CALL apply_truncation(%s)" % variable.name
    sourcefile.lines.insert(insertion_line, new_line)
    sourcefile.rebuild_text()


def insert_variable_precision_specification_to_allocatable(variable, file_objects, counter_object):
    sourcefile, allocation_line = find_file_and_line_of_allocation(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        insert_variable_precision_specification(variable, sourcefile, allocation_line + 1)
        insert_variable_usage(variable, sourcefile, allocation_line + 1)


def insert_variable_precision_specification_to_namelist_parameter(variable, file_objects, counter_object):
    sourcefile, reading_line = find_file_and_line_of_namelist_reading(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        insert_variable_precision_specification(variable, sourcefile, reading_line + 1)
        insert_variable_usage(variable, sourcefile, reading_line + 1)
    else:
        warnings.warn("Variable %s not found" % variable.name)


def insert_variable_precision_to_routine_variable(variable, file_objects, counter_object):
    sourcefile, first_use_line = find_first_use(variable, file_objects)
    if sourcefile:
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        _line = sourcefile.lines[first_use_line]
        if re.search(r'^\s*if\s*\(', _line, re.I):
            condition = close_brackets(_line)
        else:
            condition = None
        insert_variable_precision_specification(variable, sourcefile, first_use_line, condition=condition)
        insert_variable_truncation(variable, sourcefile, first_use_line + 1, condition=condition)
        insert_variable_usage(variable, sourcefile, first_use_line + 2, condition=condition)
    else:
        warnings.warn("Variable %s not found %s in routine %s" % (
            variable.name, variable.routine.module.name, variable.routine.name))


def insert_in_line(var, source_file, routine, counter, line_index, vault=None, condition=None):
    inner_index = line_index
    var_object = retrieve_variable(var, routine=routine, vault=vault)
    # Need to solve parameter issue here
    if var_object.is_parameter:
        return inner_index
    elif var_object.intent == "in":
        return inner_index

    if condition is None:
        condition = ""
    else:
        condition = condition

    sbits_assignation = "    %s%%sbits = emulator_variable_precisions(%i)"
    apply_truncation = "    CALL apply_truncation(%s) "

    if var_object.id is None:
        var_object.id = counter.count
        var_object.mutable = True
        counter.up()
    if var.count('='):
        var = var.split('=')[1].strip()

    source_file.lines.insert(inner_index + 1, condition + sbits_assignation % (var, var_object.id))
    source_file.lines.insert(inner_index + 2, condition + apply_truncation % var)

    return inner_index + 2


def add_function_to_argument(string, field_index, _function, number_of_fields=None):
    # Given a line with a call, add a call to a function in a specific field by index.
    num_arguments, interval = find_argument_intervals(string, field_index)
    if num_arguments <= field_index:
        return string
    if number_of_fields:
        if number_of_fields is not num_arguments:
            return string
    first_segment = string[:interval[0]]
    intermediate_segment = string[interval[0]:interval[1]]
    last_segment = string[interval[1]:]

    # The arguments containing a literal are discarded (Probably a bad workaround)
    if intermediate_segment.count("'"):
        #     return _line
        # raise ValueError("Probably the argument should not contain quotes at this point")
        warnings.warn("Probably the argument should not contain quotes at this point:\n%s" % intermediate_segment)

    # The arguments that already contain a call to the function are discarded (Probably a bad workaround)
    if re.match(r"^ *%s *\(.*" % _function, intermediate_segment):
        # raise ValueError("Probably the argument should not contain %s at this point" % _function)
        return string

    # If there are named arguments, add the function to the actual argument
    # (if the equal sign is inside a parenthesis, this should not be done)
    named_argument_workaround = False

    if intermediate_segment.count("="):
        if intermediate_segment.count("("):
            if intermediate_segment.find("=") < intermediate_segment.find("("):
                named_argument_workaround = True
        else:
            named_argument_workaround = True

    if named_argument_workaround:
        parts = intermediate_segment.split("=")
        intermediate_segment = "%s=%s( %s )" % (parts[0], _function, parts[1])
    else:
        intermediate_segment = "%s( %s )" % (_function, intermediate_segment)
    string = "%s%s%s" % (first_segment, intermediate_segment, last_segment)
    return string


def add_val_to_argument(string, _field):
    num_arguments, intervals = find_argument_intervals(string)
    for i in [_field]:
        num_arguments, intervals = find_argument_intervals(string)
        interval = intervals[i]
        start = interval[0]
        end = interval[1]
        if string[start] == "/":
            start += 1
        if string[end - 1] == "/":
            end -= 1
        first_segment = string[:start]
        intermediate_segment = string[start:end]
        last_segment = string[end:]
        if intermediate_segment.strip()[:2] == "(/":
            intermediate_segment = intermediate_segment.strip()[2:]
            first_segment = first_segment + "(/"
        if intermediate_segment.strip()[-2:] == "/)":
            intermediate_segment = intermediate_segment.strip()[:-2]
            last_segment = "/)" + last_segment

        string = "%s %s%%val %s" % (first_segment, intermediate_segment, last_segment)
    return string


def add_rpe_to_argument(string, field_index, number_of_fields=None):
    string = add_function_to_argument(string, field_index, "rpe", number_of_fields)
    return string


def add_rpe_var_to_arguments(string):
    num_arguments, intervals = find_argument_intervals(string)
    for i in range(num_arguments):
        num_arguments, intervals = find_argument_intervals(string)
        interval = intervals[i]
        start = interval[0]
        end = interval[1]
        if string[start] == "/":
            start += 1
        if string[end - 1] == "/":
            end -= 1
        first_segment = string[:start]
        intermediate_segment = string[start:end]
        last_segment = string[end:]
        string = "%s rpe_var( 52, %s ) %s" % (first_segment, intermediate_segment, last_segment)
    return string


def add_real_to_argument(string, field_index, number_of_fields=None):
    string = add_function_to_argument(string, field_index, "real", number_of_fields)
    return string


def replace_variable_exact_match(piece_to_replace, replacement, str):
    # In case our string does not start with a - or +, we'll add the \b to indicate a word border
    if piece_to_replace.strip()[0] not in ["-", "+"]:
        piece_to_replace = "\\b" + piece_to_replace.strip()
    # Adapt the string to fulfill regex rules
    piece_to_replace = piece_to_replace.replace("+", r"\+")
    piece_to_replace = piece_to_replace.replace("-", r"\-")
    piece_to_replace = piece_to_replace.replace("*", r"\*")
    if piece_to_replace.count("("):
        piece_to_replace = piece_to_replace.replace("(", "\s*\(")
        # Avoid matching structures
        if not piece_to_replace.count("%"):
            piece_to_replace = piece_to_replace.replace(")", "\)(?!%)")
        else:
            piece_to_replace = piece_to_replace.replace(")", "\)")
    else:
        piece_to_replace = piece_to_replace + "\\b(?!%val)"

    # Replace string using regex
    str = re.sub(piece_to_replace, replacement, str)
    return str
